import pytest
import pandas as pd
from genomicranges import GenomicRanges
from random import random
from iranges import IRanges
from biocframe import BiocFrame
import numpy as np

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"

gr = GenomicRanges(
    seqnames=[
        "chr1",
        "chr2",
        "chr3",
        "chr2",
        "chr3",
    ],
    ranges=IRanges([x for x in range(101, 106)], [11, 21, 25, 30, 5]),
    strand=["*", "-", "*", "+", "-"],
    mcols=BiocFrame(
        {
            "score": range(0, 5),
            "GC": [random() for _ in range(5)],
        }
    ),
)


def test_tile():
    assert gr is not None

    tiles = gr.tile(n=2)
    assert tiles is not None
    assert isinstance(tiles, GenomicRanges)
    assert len(tiles) == 10


def test_slide():
    assert gr is not None

    tiles = gr.sliding_windows(width=3)

    assert tiles is not None
    assert isinstance(tiles, GenomicRanges)
    assert len(tiles) == 82


def test_tile_genome():
    seqlengths = {"chr1": 100, "chr2": 75, "chr3": 200}

    tiles = GenomicRanges.tile_genome(seqlengths=seqlengths, n=10)

    assert tiles is not None
    assert isinstance(tiles, GenomicRanges)
    assert len(tiles) == 30



from itertools import groupby
from typing import List, Optional, Sequence, Tuple, Union

import biocutils as ut
import numpy as np

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"

STRAND_MAP = {"+": 1, "-": -1, "*": 0}
REV_STRAND_MAP = {"1": "+", "-1": "-", "0": "*"}


def sanitize_strand_vector(
    strand: Union[Sequence[str], Sequence[int], np.ndarray]
) -> np.ndarray:
    """Create a numpy representation for ``strand``.

    Mapping: 1 for "+" (forward strand), 0 for "*" (any strand) and -1 for "-" (reverse strand).

    Args:
        strand: List of strand.

    Raises:
        ValueError:
            If strand is None.
            If strand contains values other than +,- and *.
            If strand is not a numpy vector, string of integers or strings.

    Returns:
        A numpy vector.
    """
    if strand is None:
        raise ValueError("'strand' cannot be None.")

    if isinstance(strand, np.ndarray):
        if len(strand.shape) > 1:
            raise ValueError("'strand' must be a 1-dimensional vector.")

        if not set(np.unique(strand)).issubset([-1, 0, 1]):
            raise ValueError(
                "'strand' must only contain values 1 (forward strand), -1 (reverse strand) or 0 (reverse strand)."
            )
        return strand.astype(np.int8)
    elif ut.is_list_of_type(strand, str):
        if not set(strand).issubset(["+", "-", "*"]):
            raise ValueError("Values in 'strand' must be either +, - or *.")
        return np.asarray([STRAND_MAP[x] for x in strand], dtype=np.int8)
    elif ut.is_list_of_type(strand, (int, float)):
        if not set(strand).issubset([1, 0, -1]):
            raise ValueError(
                "'strand' must only contain values 1 (forward strand), -1 (reverse strand) or 0 (reverse strand)."
            )
        return np.asarray(strand, dtype=np.int8)
    else:
        raise TypeError(
            "'strand' must be either a numpy vector, a list of integers or strings representing strand."
        )


def _sanitize_vec(x: Sequence):
    if isinstance(x, np.ma.MaskedArray):
        x.filled(fill_value=None)
        return x.tolist()

    return list(x)


def _sanitize_strand_search_ops(query_strand, subject_strand):
    query_strand = REV_STRAND_MAP[query_strand]
    subject_strand = REV_STRAND_MAP[subject_strand]

    out = None

    if query_strand == "+":
        if subject_strand == "+":
            out = "+"
        elif subject_strand == "-":
            out = None
        elif subject_strand == "*":
            out = "+"
    elif query_strand == "-":
        if subject_strand == "+":
            out = None
        elif subject_strand == "-":
            out = "-"
        elif subject_strand == "*":
            out = "-"
    elif query_strand == "*":
        if subject_strand == "*":
            out = "+"
        elif subject_strand == "-":
            out = "-"
        elif subject_strand == "*":
            out = "-"

    if out is None:
        return None

    return STRAND_MAP[out]


def split_intervals(start: int, end: int, step: int) -> List:
    """Split an interval range into equal bins.

    Args:
        start:
            Start interval.

        end:
            End interval.

        step:
            Width or step of each interval.

    Returns:
        List of intervals split into bins.
    """
    bins = []
    for i in range(start, end + 1, step):
        bins.append((i, min(i + step - 1, end) - i))

    return bins


def slide_intervals(start: int, end: int, width: int, step: int) -> List:
    """Sliding intervals.

    Args:
        start:
            Start interval.

        end:
            End interval.

        step:
            Step of each interval.

        width:
            Width of each interval.

    Returns:
        List of intervals split into bins.
    """
    bins = []

    if end - width < start:
        bins.append((start, end - start))
    else:
        for i in range(start, end - width + 2, step):
            bins.append((i, min(i + width - 1, end) - i))

    return bins


def create_np_vector(
    intervals: List[Tuple[int, int]],
    with_reverse_map: bool = False,
    force_size: Optional[int] = None,
    dont_sum: bool = False,
    value: int = 1,
) -> Tuple[np.ndarray, Optional[List]]:
    """Represent intervals and calculate coverage.

    Args:
        intervals:
            Input interval vector.

        with_reverse_map:
            Return map of indices? Defaults to False.

        force_size:
            Force size of the array.

        dont_sum:
            Do not sum. Defaults to False.

        value:
            Default value to increment. Defaults to 1.

    Returns:
        A numpy array representing coverage from the
        intervals and optionally a reverse index map.
    """
    if len(intervals) < 1:
        return intervals

    max_end = force_size
    if max_end is None:
        max_end = max([x[1] for x in intervals])
    cov = np.zeros(max_end)

    revmap = None
    if with_reverse_map:
        revmap = [[] for _ in range(max_end)]

    for idx in range(len(intervals)):
        i = intervals[idx]

        if dont_sum:
            cov[i[0] - 1 : i[1]] = value
        else:
            cov[i[0] - 1 : i[1]] += value

        if with_reverse_map:
            _ = [revmap[x].append(idx + 1) for x in range(i[0] - 1, i[1])]

    return cov, revmap


def group_by_indices(groups: list) -> dict:
    return {
        k: [x[0] for x in v]
        for k, v in groupby(
            sorted(enumerate(groups), key=lambda x: x[1]), lambda x: x[1]
        )
    }



import pytest
import pandas as pd
from genomicranges.GenomicRanges import GenomicRanges
from random import random
from iranges import IRanges
from biocframe import BiocFrame
import numpy as np

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"

gr = GenomicRanges(
    seqnames=[
        "chr1",
        "chr2",
        "chr3",
        "chr2",
        "chr3",
    ],
    ranges=IRanges([x for x in range(101, 106)], [11, 21, 25, 30, 5]),
    strand=["*", "-", "*", "+", "-"],
    mcols=BiocFrame(
        {
            "score": range(0, 5),
            "GC": [random() for _ in range(5)],
        }
    ),
)


def test_matches():
    assert gr is not None

    query_hits = gr.match(gr[2:5])

    assert query_hits is not None
    assert query_hits == [[2], [3], [4]]


def test_order():
    assert gr is not None

    order = gr.order()
    assert (order == np.array([0, 1, 3, 4, 2])).all()


def test_sort():
    assert gr is not None
    result = gr.sort()

    assert result is not None

    result = gr.sort(decreasing=True)
    assert result is not None


def test_rank():
    assert gr is not None
    result = gr.rank()

    assert result is not None
    assert (result == np.array([0, 1, 4, 2, 3])).all()



from typing import Dict, List, Optional, Sequence, Union
from warnings import warn

import biocutils as ut

from .utils import _sanitize_vec

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


def _validate_seqnames(seqnames):
    if not ut.is_list_of_type(seqnames, str):
        raise ValueError("'seqnames' should be a list of strings.")

    n = len(seqnames)
    if n != len(set(seqnames)):
        raise ValueError("'seqnames' should contain unique strings.")


def _validate_seqlengths(seqlengths, num_seqs):
    if not ut.is_list_of_type(seqlengths, int, ignore_none=True):
        raise ValueError("'seqlengths' should be a list of integers.")

    if num_seqs != len(seqlengths):
        raise ValueError("'seqnames' and 'seqlengths' should have the same length.")

    for sl in seqlengths:
        if sl is not None and sl < 0:
            raise ValueError("all entries of 'seqlengths' should be non-negative.")


def _validate_is_circular(is_circular, num_seqs):
    if not ut.is_list_of_type(is_circular, bool, ignore_none=True):
        raise ValueError("'is_circular' should be a list of booleans.")

    if num_seqs != len(is_circular):
        raise ValueError("'seqnames' and 'is_circular' should have the same length.")


def _validate_genome(genome, num_seqs):
    if not ut.is_list_of_type(genome, str, ignore_none=True):
        raise ValueError("'genome' should be a list of strings.")

    if num_seqs != len(genome):
        raise ValueError("'seqnames' and 'genome' should have the same length.")


class SeqInfoIterator:
    """An iterator to a :py:class:`~SeqInfo` object."""

    def __init__(self, obj: "SeqInfo") -> None:
        """Initialize the iterator.

        Args:
            obj:
                Source object to iterate.
        """
        self._sinfo = obj
        self._current_index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._current_index < len(self._sinfo):
            iter_row_index = self._sinfo._seqnames[self._current_index]

            iter_slice = self._sinfo[self._current_index]
            self._current_index += 1
            return (iter_row_index, iter_slice)

        raise StopIteration


class SeqInfo:
    """Information about the reference sequences, specifically the name and length of each sequence, whether it is a
    circular, and the identity of the genome from which it was derived."""

    def __init__(
        self,
        seqnames: Sequence[str],
        seqlengths: Optional[Union[int, Sequence[int], Dict[str, int]]] = None,
        is_circular: Optional[Union[bool, Sequence[bool], Dict[str, bool]]] = None,
        genome: Optional[Union[str, Sequence[str], Dict[str, str]]] = None,
        validate: bool = True,
    ) -> None:
        """
        Args:
            seqnames:
                Names of all reference sequences, should be unique.

            seqlengths:
                Lengths of all sequences in base pairs. This should contain
                non-negative values and have the same number of elements as
                ``seqnames``. Entries may also be None if no lengths are
                available for that sequence.

                Alternatively, a dictionary where keys are the sequence names
                and values are the lengths. If a name is missing from this
                dictionary, the length of the sequence is set to None.

                Alternatively a single integer, if all sequences are of the
                same length.

                Alternatively None, if no length information is available
                for any sequence.

            is_circular:
                Whether each sequence is circular. This should have the same
                number of elements as ``seqnames``. Entries may also be None
                if no information is available for that sequence.

                Alternatively, a dictionary where keys are the sequence names
                and values are the circular flags. If a name is missing from
                this dictionary, the flag for the sequence is set to None.

                Alternatively a single boolean, if all sequences have the same
                circular flag.

                Alternatively None, if no flags are available for any sequence.

            genome:
                The genome build containing each reference sequence. This
                should have the same number of elements as ``seqnames``.
                Entries may also be None if no information is available.

                Alternatively, a dictionary where keys are the sequence names
                and values are the genomes. If a name is missing from this
                dictionary, the genome is set to None.

                Alternatively a single string, if all sequences are derived
                from the same genome.

                Alternatively None, if no genome information is available
                for any sequence.

            validate:
                Whether to validate the arguments, internal use only.
        """
        self._seqnames = list(seqnames)
        self._reverse_seqnames = None
        self._seqlengths = self._flatten_incoming(seqlengths, int)
        self._is_circular = self._flatten_incoming(is_circular, bool)
        self._genome = self._flatten_incoming(genome, str)

        if validate:
            _validate_seqnames(self._seqnames)
            num_seqs = len(self._seqnames)
            _validate_seqlengths(self._seqlengths, num_seqs)
            _validate_is_circular(self._is_circular, num_seqs)
            _validate_genome(self._genome, num_seqs)

    def _populate_reverse_seqnames_index(self):
        if self._reverse_seqnames is None:
            revmap = {}
            for i, n in enumerate(self._seqnames):
                if n not in revmap:
                    revmap[n] = i
            self._reverse_seqnames = revmap

    def _wipe_reverse_seqnames_index(self):
        self._reverse_seqnames = None

    def _flatten_incoming(self, values, expected) -> List:
        if values is None or isinstance(values, expected):
            return [values] * len(self)

        if isinstance(values, dict):
            output = []
            for n in self._seqnames:
                if n in values:
                    output.append(values[n])
                else:
                    output.append(None)
            return output

        values = _sanitize_vec(values)

        if isinstance(values, list):
            return values

        return list(values)

    def _define_output(self, in_place: bool = False) -> "SeqInfo":
        if in_place is True:
            return self
        else:
            return self.__copy__()

    #########################
    ######>> Copying <<######
    #########################

    def __deepcopy__(self, memo=None, _nil=[]):
        """
        Returns:
            A deep copy of the current ``SeqInfo``.
        """
        from copy import deepcopy

        _seqnames_copy = deepcopy(self._seqnames)
        _seqlengths_copy = deepcopy(self._seqlengths)
        _is_circular_copy = deepcopy(self._is_circular)
        _genome_copy = deepcopy(self._genome)

        current_class_const = type(self)
        return current_class_const(
            seqnames=_seqnames_copy,
            seqlengths=_seqlengths_copy,
            is_circular=_is_circular_copy,
            genome=_genome_copy,
            validate=False,
        )

    def __copy__(self):
        """
        Returns:
            A shallow copy of the current ``SeqInfo``.
        """
        current_class_const = type(self)
        return current_class_const(
            self._seqnames,
            self._seqlengths,
            self._is_circular,
            self._genome,
            validate=False,
        )

    def copy(self):
        """Alias for :py:meth:`~__copy__`."""
        return self.__copy__()

    ##########################
    ######>> Printing <<######
    ##########################

    def __repr__(self) -> str:
        """
        Returns:
            A string representation of this ``SeqInfo``.
        """
        output = "SeqInfo(number_of_seqnames=" + str(len(self))
        output += ", seqnames=" + ut.print_truncated_list(self._seqnames)
        output += ", seqlengths=" + repr(self._seqlengths)
        output += ", is_circular=" + ut.print_truncated_list(self._is_circular)
        output += ", genome=" + ut.print_truncated_list(self._genome)

        output += ")"
        return output

    def __str__(self) -> str:
        """
        Returns:
            A pretty-printed string containing the contents of this ``SeqInfo``.
        """
        output = f"SeqInfo with {len(self)} sequence{'s' if len(self) != 1 else ''}\n"

        nr = len(self)
        added_table = False
        if nr:
            if nr <= 10:
                indices = range(nr)
                insert_ellipsis = False
            else:
                indices = [0, 1, 2, nr - 3, nr - 2, nr - 1]
                insert_ellipsis = True

            raw_floating = ut.create_floating_names(None, indices)
            if insert_ellipsis:
                raw_floating = raw_floating[:3] + [""] + raw_floating[3:]
            floating = ["", ""] + raw_floating

            columns = []

            header = ["seqnames", "<str>"]
            showed = [f"{self._seqnames[x]}" for x in indices]
            if insert_ellipsis:
                showed = showed[:3] + ["..."] + showed[3:]
            columns.append(header + showed)

            header = ["seqlengths", f"<{ut.print_type(self._seqlengths)}>"]
            showed = [f"{self._seqlengths[x]}" for x in indices]
            if insert_ellipsis:
                showed = showed[:3] + ["..."] + showed[3:]
            columns.append(header + showed)

            header = ["is_circular", f"<{ut.print_type(self._is_circular)}>"]
            showed = [f"{self._is_circular[x]}" for x in indices]
            if insert_ellipsis:
                showed = showed[:3] + ["..."] + showed[3:]
            columns.append(header + showed)

            header = ["genome", f"<{ut.print_type(self._genome)}>"]
            showed = [f"{self._genome[x]}" for x in indices]
            if insert_ellipsis:
                showed = showed[:3] + ["..."] + showed[3:]
            columns.append(header + showed)

            output += ut.print_wrapped_table(columns, floating_names=floating)
            added_table = True

        footer = []
        if len(footer):
            if added_table:
                output += "\n------\n"
            output += "\n".join(footer)

        return output

    ##########################
    ######>> seqnames <<######
    ##########################

    def get_seqnames(self) -> List[str]:
        """
        Returns:
            List of all chromosome names.
        """
        return self._seqnames

    def set_seqnames(
        self, seqnames: Sequence[str], in_place: bool = False
    ) -> "SeqInfo":
        """
        Args:
            seqnames:
                List of sequence names, of length equal to the number of names
                in this ``SeqInfo`` object. All names should be unique strings.

            in_place:
                Whether to modify the ``SeqInfo`` object in place.

        Returns:
            A modified ``SeqInfo`` object, either as a copy of the original
            or as a reference to the (in-place-modified) original.
        """

        _validate_seqnames(list(seqnames))

        output = self._define_output(in_place)
        output._seqnames = list(seqnames)
        return output

    @property
    def seqnames(self) -> List[str]:
        warn("'seqnames' is deprecated, use 'get_seqnames' instead", UserWarning)
        return self.get_seqnames()

    @seqnames.setter
    def seqnames(self, seqnames: Sequence[str]):
        warn(
            "Setting property 'seqnames' is an in-place operation, use 'set_seqnames' instead",
            UserWarning,
        )

        self.set_seqnames(seqnames, in_place=True)

    ############################
    ######>> seqlengths <<######
    ############################

    def get_seqlengths(self) -> List[int]:
        """
        Returns:
            A list of integers is returned containing the lengths of all
            sequences, in the same order as the sequence names from
            :py:meth:`~get_seqnames`.
        """
        return self._seqlengths

    def set_seqlengths(
        self,
        seqlengths: Optional[Union[int, Sequence[int], Dict[str, int]]],
        in_place: bool = False,
    ) -> "SeqInfo":
        """
        Args:
            seqlengths:
                List of sequence lengths, of length equal to the number of
                names in this ``SeqInfo`` object. Values may be None or
                non-negative integers.

                Alternatively, a dictionary where keys are the sequence
                names and values are the lengths. Not all names need to be
                present in which case the length is assumed to be None.

            in_place:
                Whether to modify the ``SeqInfo`` object in place.

        Returns:
            A modified ``SeqInfo`` object, either as a copy of the original
            or as a reference to the (in-place-modified) original.
        """
        _seqlengths = self._flatten_incoming(seqlengths, int)
        _validate_seqlengths(_seqlengths, len(self))

        output = self._define_output(in_place)
        output._seqlengths = _seqlengths
        return output

    @property
    def seqlengths(self) -> List[int]:
        warn(
            "'seqlengths' is deprecated, use 'get_seqlengths' instead",
            UserWarning,
        )
        return self.get_seqlengths()

    @seqlengths.setter
    def seqlengths(
        self, seqlengths: Optional[Union[int, Sequence[int], Dict[str, int]]]
    ):
        warn(
            "Setting property 'seqlengths' is an in-place operation, use 'set_seqlengths' instead",
            UserWarning,
        )

        self.set_seqlengths(seqlengths, in_place=True)

    #############################
    ######>> is-circular <<######
    #############################

    def get_is_circular(self) -> List[bool]:
        """
        Returns:
            A list of booleans is returned specifying whether each sequence
            (from :py:meth:`~get_seqnames`) is circular.
        """
        return self._is_circular

    def set_is_circular(
        self,
        is_circular: Optional[Union[bool, Sequence[bool], Dict[str, bool]]],
        in_place: bool = False,
    ) -> "SeqInfo":
        """
        Args:
            is_circular:
                List of circular flags, of length equal to the number of
                names in this ``SeqInfo`` object. Values may be None or
                booleans.

                Alternatively, a dictionary where keys are the sequence
                names and values are the flags. Not all names need to be
                present in which case the flag is assumed to be None.

            in_place:
                Whether to modify the ``SeqInfo`` object in place.

        Returns:
            A modified ``SeqInfo`` object, either as a copy of the original
            or as a reference to the (in-place-modified) original.
        """

        _is_circular = self._flatten_incoming(is_circular, bool)
        _validate_is_circular(_is_circular, len(self))

        output = self._define_output(in_place)
        output._is_circular = _is_circular
        return output

    @property
    def is_circular(self) -> List[bool]:
        warn(
            "'is_circular' is deprecated, use 'get_is_circular' instead",
            UserWarning,
        )
        return self.get_is_circular()

    @is_circular.setter
    def is_circular(
        self, is_circular: Optional[Union[bool, Sequence[bool], Dict[str, bool]]]
    ):
        warn(
            "Setting property 'is_circular' is an in-place operation, use 'set_is_circular' instead",
            UserWarning,
        )

        self.set_is_circular(is_circular, in_place=True)

    ########################
    ######>> genome <<######
    ########################

    def get_genome(self) -> List[str]:
        """
        Returns:
            A list of strings is returned containing the genome identity for
            all sequences in :py:meth:`~get_seqnames`.
        """
        return self._genome

    def set_genome(
        self,
        genome: Optional[Union[str, Sequence[str], Dict[str, str]]],
        in_place: bool = False,
    ) -> "SeqInfo":
        """
        Args:
            genome:
                List of genomes, of length equal to the number of names in this
                ``SeqInfo`` object. Values may be None or strings.


            in_place:
                Whether to modify the ``SeqInfo`` object in place.

        Returns:
            A modified ``SeqInfo`` object, either as a copy of the original
            or as a reference to the (in-place-modified) original.
        """
        _genome = self._flatten_incoming(genome, str)
        _validate_genome(_genome, len(self))

        output = self._define_output(in_place)
        output._genome = _genome
        return output

    @property
    def genome(self) -> List[str]:
        warn("'genome' is deprecated, use 'get_genome' instead", UserWarning)
        return self.get_genome()

    @genome.setter
    def genome(self, genome: Optional[Union[bool, Sequence[bool], Dict[str, bool]]]):
        warn(
            "Setting property 'genome' is an in-place operation, use 'set_genome' instead",
            UserWarning,
        )

        self.set_genome(genome, in_place=True)

    ######################################
    ######>> length and iterators <<######
    ######################################

    def __len__(self) -> int:
        """
        Returns:
            Number of sequences in this object.
        """
        return len(self._seqnames)

    def __iter__(self) -> SeqInfoIterator:
        """Iterator over sequences."""
        return SeqInfoIterator(self)

    #########################
    ######>> Slicers <<######
    #########################

    def get_subset(self, subset: Union[str, int, bool, Sequence]) -> "SeqInfo":
        """Subset ``SeqInfo``, based on their indices or seqnames.

        Args:
            subset:
                Indices to be extracted. This may be an integer, boolean, string,
                or any sequence thereof, as supported by
                :py:meth:`~biocutils.normalize_subscript.normalize_subscript`.
                Scalars are treated as length-1 sequences.

                Strings may only be used if :py:attr:``~seqnames`` are available (see
                :py:meth:`~get_seqnames`). The first occurrence of each string
                in the seqnames is used for extraction.

        Returns:
            A new ``SeqInfo`` object with the sequences of interest.
        """
        if len(self) == 0:
            return SeqInfo.empty()

        idx, _ = ut.normalize_subscript(subset, len(self), self._seqnames)

        current_class_const = type(self)
        return current_class_const(
            seqnames=ut.subset_sequence(self._seqnames, idx),
            seqlengths=ut.subset_sequence(self._seqlengths, idx),
            is_circular=ut.subset_sequence(self._is_circular, idx),
            genome=ut.subset_sequence(self._genome, idx),
        )

    def __getitem__(self, subset: Union[str, int, bool, Sequence]) -> "SeqInfo":
        """Alias to :py:attr:`~get_subset`."""
        return self.get_subset(subset)

    @classmethod
    def empty(cls):
        """Create an zero-length `SeqInfo` object.

        Returns:
            same type as caller, in this case a `SeqInfo`.
        """
        return SeqInfo([], [], [], [])


@ut.combine_sequences.register
def _combine_SeqInfo(*x: SeqInfo) -> SeqInfo:
    return merge_SeqInfo(x)


def merge_SeqInfo(objects: List[SeqInfo]) -> SeqInfo:
    """Merge multiple :py:class:`~SeqInfo` objects, taking the union of all reference sequences. If the same reference
    sequence is present with the same details across ``objects``, only a single instance is present in the final object;
    if details are contradictory, they are replaced with None.

    Args:
        objects: List of ``SeqInfo`` objects.

    Returns:
        A single merged ``SeqInfo`` object.
    """
    all_sequences = {}

    for obj in objects:
        for i, y in enumerate(obj._seqnames):
            curlen = obj._seqlengths[i]
            curcir = obj._is_circular[i]
            curgen = obj._genome[i]

            if y not in all_sequences:
                all_sequences[y] = [curlen, curcir, curgen]
            else:
                present = all_sequences[y]
                prelen, precir, pregen = present
                if prelen != curlen:
                    present[0] = None
                if precir != curcir:
                    present[1] = None
                if pregen != curgen:
                    present[2] = None

    out_names = []
    out_lengths = []
    out_circular = []
    out_genome = []
    for k, v in all_sequences.items():
        out_names.append(k)
        out_lengths.append(v[0])
        out_circular.append(v[1])
        out_genome.append(v[2])

    return SeqInfo(out_names, out_lengths, out_circular, out_genome, validate=False)



import pytest
import pandas as pd
from genomicranges import GenomicRanges, GenomicRangesList
from biocframe import BiocFrame
from iranges import IRanges
from random import random
import genomicranges

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"

a = GenomicRanges(
    seqnames=["chr1", "chr2", "chr1", "chr3"],
    ranges=IRanges([1, 3, 2, 4], [10, 30, 50, 60]),
    strand=["-", "+", "*", "+"],
    mcols=BiocFrame({"score": [1, 2, 3, 4]}),
)

b = GenomicRanges(
    seqnames=["chr2", "chr4", "chr5"],
    ranges=IRanges([3, 6, 4], [30, 50, 60]),
    strand=["-", "+", "*"],
    mcols=BiocFrame({"score": [2, 3, 4]}),
)


def test_create_grl():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    assert isinstance(grl, GenomicRangesList)
    assert isinstance(grl["a"], GenomicRanges)
    assert grl["b"] is not None
    assert len(grl) == 2


def test_create_grl_should_fail():
    with pytest.raises(Exception):
        GenomicRangesList(ranges=[a, 2])


def test_empty_grl():
    grl = GenomicRangesList.empty(n=100)
    assert isinstance(grl, GenomicRangesList)


def test_grl_set_names():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    grl_replace_names = grl.set_names(["aa", "bb"])

    assert grl_replace_names is not None
    assert isinstance(grl_replace_names, GenomicRangesList)
    assert list(grl.get_names()) == ["a", "b"]



import pytest
import pandas as pd
from genomicranges.GenomicRanges import GenomicRanges
from random import random
from iranges import IRanges
from biocframe import BiocFrame
import numpy as np

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"

gr = GenomicRanges(
    seqnames=[
        "chr1",
        "chr2",
        "chr3",
        "chr2",
        "chr3",
    ],
    ranges=IRanges([x for x in range(101, 106)], [11, 21, 25, 30, 5]),
    strand=["*", "-", "*", "+", "-"],
    mcols=BiocFrame(
        {
            "score": range(0, 5),
            "GC": [random() for _ in range(5)],
        }
    ),
)


def test_invert():
    assert gr is not None

    result = gr.invert_strand()

    assert result is not None
    assert len(result._strand) == len(gr._strand)


def test_sample():
    assert gr is not None

    result = gr.sample(k=3)

    assert result is not None
    assert len(result) == 3



import pytest
import pandas as pd
from genomicranges import GenomicRanges, GenomicRangesList
from biocutils import combine_sequences
from biocframe import BiocFrame
from iranges import IRanges
from random import random
import genomicranges

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"

a = GenomicRanges(
    seqnames=["chr1", "chr2", "chr1", "chr3"],
    ranges=IRanges([1, 3, 2, 4], [10, 30, 50, 60]),
    strand=["-", "+", "*", "+"],
    mcols=BiocFrame({"score": [1, 2, 3, 4]}),
)

b = GenomicRanges(
    seqnames=["chr2", "chr4", "chr5"],
    ranges=IRanges([3, 6, 4], [30, 50, 60]),
    strand=["-", "+", "*"],
    mcols=BiocFrame({"score": [2, 3, 4]}),
)


def test_is_empty_False():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    assert grl.is_empty() == False


def test_is_empty_slice():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    assert grl.is_empty() == False

    sgrl = grl[0:1]
    assert sgrl is not None
    assert isinstance(sgrl, GenomicRangesList)
    assert len(sgrl) == 1


def test_slice_by_name():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    assert grl.is_empty() == False

    sgrl = grl[["a"]]
    assert sgrl is not None
    assert isinstance(sgrl, GenomicRangesList)
    assert len(sgrl) == 1


def test_slice_by_bool():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    assert grl.is_empty() == False

    sgrl = grl[[True, False]]
    assert sgrl is not None
    assert isinstance(sgrl, GenomicRangesList)
    assert len(sgrl) == 1

    with pytest.raises(Exception):
        grl[[False]]


def test_is_empty_True():
    grl = GenomicRangesList(GenomicRanges.empty(), range_lengths=[0])

    assert grl.is_empty() == True
    assert len(grl) == 1


def test_nrows():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    nrows = grl.element_nrows()
    assert isinstance(nrows, dict)
    assert list(nrows.keys()) == ["a", "b"]
    assert list(nrows.values()) == [4, 3]


def test_props():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    props = ["start", "end", "strand"]

    for prop in props:
        v = getattr(grl, prop)

        assert isinstance(v, dict)

    assert isinstance(grl.mcols, BiocFrame)


def test_to_pandas():
    grl = GenomicRangesList(ranges=[a, b], names=["a", "b"])

    df = grl.to_pandas()
    assert len(df) == 7
    assert df.index.tolist() == [
        "a",
        "a",
        "a",
        "a",
        "b",
        "b",
        "b",
    ]


def test_combine():
    grla = GenomicRangesList(ranges=[a], names=["a"])
    grlb = GenomicRangesList(ranges=[b, a], names=["b", "c"])

    cgrl = combine_sequences(grla, grlb)

    assert len(cgrl) == 3


def test_empty_grl_slice():
    grl = GenomicRangesList.empty(n=100)
    assert isinstance(grl, GenomicRangesList)

    subset = grl[0:10]
    assert isinstance(subset, GenomicRangesList)
    assert len(subset) == 10

    subset = grl[[1, 2, 3]]
    assert isinstance(subset, GenomicRangesList)
    assert len(subset) == 3



import pytest
from genomicranges import GenomicRanges
from iranges import IRanges
from biocframe import BiocFrame
from random import random
import pandas as pd

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"


def test_create_gr():
    gr = GenomicRanges(
        seqnames=[
            "chr1",
            "chr2",
            "chr2",
            "chr2",
            "chr1",
            "chr1",
            "chr3",
            "chr3",
            "chr3",
            "chr3",
        ],
        ranges=IRanges(start=range(100, 110), width=range(110, 120)),
        strand=["-", "+", "+", "*", "*", "+", "+", "+", "-", "-"],
        mcols=BiocFrame(
            {
                "score": range(0, 10),
                "GC": [random() for _ in range(10)],
            }
        ),
    )

    assert gr is not None
    assert len(gr) == 10


def test_gr_empty():
    gre = GenomicRanges.empty()

    assert gre is not None
    assert isinstance(gre, GenomicRanges)
    assert len(gre) == 0



import pytest
import pandas as pd
from genomicranges.GenomicRanges import GenomicRanges
from random import random
from iranges import IRanges
from biocframe import BiocFrame
import numpy as np

__author__ = "jkanche"
__copyright__ = "jkanche"
__license__ = "MIT"

gr0 = GenomicRanges(
    seqnames=["chr1", "chr1"], ranges=IRanges([2, 9], [6, 11]), strand=["+", "-"]
)

gr1 = GenomicRanges(seqnames=["chr1"], ranges=IRanges([5], [6]), strand=["-"])


def test_union():
    gr0 = GenomicRanges(
        seqnames=["chr1", "chr1"], ranges=IRanges([2, 9], [6, 11]), strand=["+", "-"]
    )
    gr1 = GenomicRanges(seqnames=["chr1"], ranges=IRanges([5], [6]), strand=["-"])

    assert gr0 is not None
    assert gr1 is not None

    out = gr0.union(gr1)

    assert out is not None
    assert out.seqnames == ["chr1", "chr1"]
    assert (out.start == np.array([2, 5])).all()
    assert (out.width == np.array([6, 15])).all()
    assert (out.strand == np.array([1, -1])).all()


def test_diff():
    gr0 = GenomicRanges(
        seqnames=["chr1", "chr1"], ranges=IRanges([2, 9], [6, 11]), strand=["+", "-"]
    )

    gr1 = GenomicRanges(seqnames=["chr1"], ranges=IRanges([5], [6]), strand=["-"])

    assert gr0 is not None
    assert gr1 is not None

    out = gr0.setdiff(gr1)

    assert out is not None
    assert out.seqnames == ["chr1", "chr1"]
    assert (out.start == np.array([2, 11])).all()
    assert (out.width == np.array([6, 9])).all()
    assert (out.strand == np.array([1, -1])).all()


def test_intersect():
    gr0 = GenomicRanges(
        seqnames=["chr1", "chr1"], ranges=IRanges([2, 9], [6, 11]), strand=["+", "-"]
    )

    gr1 = GenomicRanges(seqnames=["chr1"], ranges=IRanges([5], [6]), strand=["-"])
    assert gr0 is not None
    assert gr1 is not None

    out = gr0.intersect(gr1)

    assert out is not None
    assert out.seqnames == ["chr1"]
    assert (out.start == np.array([9])).all()
    assert (out.width == np.array([2])).all()
    assert (out.strand == np.array([-1])).all()


def test_intersect_complex():
    g_src = GenomicRanges(
        seqnames=["chr1", "chr2", "chr1", "chr3", "chr2"],
        ranges=IRanges(
            start=[101, 102, 103, 104, 109], width=[112, 103, 128, 134, 111]
        ),
        strand=["*", "-", "*", "+", "-"],
    )

    g_tgt = GenomicRanges(
        seqnames=[
            "chr1",
            "chr2",
            "chr2",
            "chr2",
            "chr1",
            "chr1",
            "chr3",
            "chr3",
            "chr3",
            "chr3",
        ],
        ranges=IRanges(start=range(101, 111), width=range(121, 131)),
        strand=["*", "-", "-", "*", "*", "+", "+", "+", "-", "-"],
        mcols=BiocFrame(
            {
                "score": range(0, 10),
                "GC": [random() for _ in range(10)],
            }
        ),
    )
    assert g_src is not None
    assert g_tgt is not None

    out = g_src.intersect(g_tgt)

    assert out is not None
    assert len(out) == 3


def test_intersect_ncls():
    gr0 = GenomicRanges(
        seqnames=["chr1", "chr1"], ranges=IRanges([2, 9], [6, 11]), strand=["+", "-"]
    )

    gr1 = GenomicRanges(seqnames=["chr1"], ranges=IRanges([5], [6]), strand=["-"])
    assert gr0 is not None
    assert gr1 is not None

    out = gr0.intersect_ncls(gr1).reduce()

    assert out is not None
    assert out.seqnames == ["chr1"]
    assert (out.start == np.array([9])).all()
    assert (out.width == np.array([2])).all()
    assert (out.strand == np.array([-1])).all()


def test_diff_with_reduce():
    g = GenomicRanges(["A"], IRanges([0], [10])).reduce()
    res = g.setdiff(g)

    assert res is not None
    assert isinstance(res, GenomicRanges)
